/* badram.c: tries to make some memory addresses inaccessible.
   Copyright (C) 2006 Blaise Gassend <blaise at gassend doc com>

   This file contains ideas and modified parts from badram,
   written by Rick van Rein (C) 2001

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/mmzone.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/moduleparam.h>

#define MAGIC ((unsigned long) 0xDEADADD)

#define phys_to_page(x)          (pfn_to_page((x) >> PAGE_SHIFT))

//int badaddr = 0x13df4f10;
//#define badpage (page_to_pfn(phys_to_page(badaddr)))
//int badaddr = 0x0901000;

static inline unsigned long page_order(struct page *page) {
          return page->u.private;
}

static inline void set_page_order(struct page *page, int order) {
        page->u.private = order;
        __SetPagePrivate(page);
}

static inline void rmv_page_order(struct page *page)
{
        __ClearPagePrivate(page);
        page->u.private = 0;
}

static int bad_range(struct zone *zone, struct page *page)
{
        if (page_to_pfn(page) >= zone->zone_start_pfn + zone->spanned_pages)
                return 1;
        if (page_to_pfn(page) < zone->zone_start_pfn)
                return 1;
#ifdef CONFIG_HOLES_IN_ZONE
        if (!pfn_valid(page_to_pfn(page)))
                return 1;
#endif
        if (zone != page_zone(page))
                return 1;
        return 0;
}

static inline int page_is_buddy(struct page *page, int order)
{
       if (PagePrivate(page)           &&
           (page_order(page) == order) &&
           !PageReserved(page)         &&
            page_count(page) == 0)
               return 1;
       return 0;
}

static inline int page_is_bad(struct page *page)
{
  if (PagePrivate(page)      && 
      page->u.private == MAGIC &&
      PageReserved(page))
    return 1;
  return 0;
}

static void print_page(struct page *pg, int order)
{
  printk(KERN_ERR "page %5lx: %03i %03i %8x ", page_to_pfn(pg), page_count(pg), 
      page_mapcount(pg), (unsigned int) pg->flags);

  if (PageActive(pg)) printk("Act ");
  if (PageDirty(pg)) printk("Dirt ");
  if (PageError(pg)) printk("Err ");
  if (PageLocked(pg)) printk("Lock ");
  if (PageLRU(pg)) printk("LRU ");
  if (PageReferenced(pg)) printk("Ref ");
  if (PageReserved(pg)) printk("Res ");
  if (PageSlab(pg)) printk("Slab ");
  if (PageUptodate(pg)) printk("UpTDat ");
  if (PagePrivate(pg)) printk("Priv(%lx) ", pg->u.private);
  if (page_is_buddy(pg, order)) printk("Bud ");
  
  printk("\n");
}

static inline void print_buddies(struct page *page)
{
  int order;
  unsigned long size = 1;
  struct page *cur_page;

  for (order = 0; order < MAX_ORDER; order++, size <<= 1)
  {
    cur_page = pfn_to_page((page_to_pfn(page) & ~(size - 1)) ^ size);
    print_page(cur_page, order);
  } 
} 

/* 
 * Free all but one page in a buddy allocator block. Assumes that
 * A lock is already held on the zone.
 */

static inline void expand(struct zone *zone, struct page *page, int max_order)
{
  int order;
  struct free_area *area = zone->free_area;
  unsigned long size = 1;
  struct page *cur_page;

  /*printk(KERN_ERR "freeing page %lx order %i\n", page_to_pfn(page), 0);
  list_add(&page->lru, &area->free_list);
  area->nr_free++;
  set_page_order(page, 0);
  printk(KERN_ERR "page %lx is %i", page_to_pfn(page), PagePrivate(page));
  */

  for (order = 0; order < max_order; order++, area++, size <<= 1)
  {
    cur_page = pfn_to_page((page_to_pfn(page) & ~(size - 1)) ^ size);
/*    printk(KERN_ERR "freeing page %lx order %i %lx %lx\n", 
	page_to_pfn(cur_page), order, 
	page_to_pfn(cur_page) ^ page_to_pfn(page), 
	page_to_pfn(cur_page) & (size - 1));*/
    list_add(&cur_page->lru, &area->free_list);
    area->nr_free++;
    set_page_order(cur_page, order);
  } 
} 

/* 
 * Allocate the specific page from the buddy allocator. The page should be 
 * part of the buddy allocator block that is passed as an argument. Assumes
 * zone lock already held.
 */

static inline void allocate_specific_page(struct zone *zone, unsigned int order,
  struct page *page, struct page *pageblk)
{
  struct free_area *area = zone->free_area + order;
/*  printk(KERN_ERR "allocating page %lx parent %lx order %i zone %lx\n", 
      page_to_pfn(page), page_to_pfn(pageblk), order, (unsigned long) zone);
*/
  list_del(&pageblk->lru);
  rmv_page_order(page);
  area->nr_free--;
  zone->free_pages -= 1UL << order;
  expand(zone, page, order);
}

/* 
 * Allocate the specific page from the buddy allocator. 
 */

static int mark_page_bad(struct page *target_page)
{
  unsigned long flags;
  int cur_order;
  struct page *page;
  struct zone *zone = page_zone(target_page);
 
  printk(KERN_ERR "userbadram: trying to allocate page: %lx -> ", 
      page_to_pfn(target_page));

  /*  struct zonelist zl = { {zone_table[1], NULL} };
  page = __alloc_pages(0, 0, &zl);
  printk(KERN_ERR "allocated page: %lx\n", (unsigned long) page);
  */
  
  if (page_to_pfn(target_page) >= max_mapnr)
  {
    printk("page out of range\n");
    return 1;
  }
  
  if (page_is_bad(target_page))
  {
    printk("already marked as bad\n");
    return 0;
  }
  
  if (page_count(target_page) > 0)
  {
    printk("already allocated\n");
    return 1;
  }
    
  if (PageReserved(target_page))
  {
    printk("already reserved\n");
    print_page(target_page, -1);
    return 1;
  }
  
  spin_lock_irqsave(&zone->lock, flags);
  for (cur_order = 0; cur_order < MAX_ORDER; cur_order++)
  {
    page = pfn_to_page(page_to_pfn(target_page) & (-1 << (cur_order)));

    if (bad_range(zone, page))
      break;
/*    printk(KERN_ERR "page: %lx %i %s(%lu) ord %i\n", page_to_pfn(page), 
	page_count(page),
	PagePrivate(page) ? "P" : "NP", page->private, cur_order);*/
    
    if (page_is_buddy(page, cur_order))
    {
      allocate_specific_page(zone, cur_order, target_page, page);
      SetPageReserved(target_page);
      SetPagePrivate(target_page);
      target_page->u.private = MAGIC;
      printk("success\n");
      break;
    }
  }
  spin_unlock_irqrestore(&zone->lock, flags);
  if (cur_order == MAX_ORDER)
  {
    printk("free parent not found (abnormal situation)");
    return 2;
  }
  
  return 0;
} 

static void print_zone(struct zone *zone)
{
  unsigned long flags;
  int current_order;
  struct free_area *area;
  struct page *page;
  int i;
  unsigned long order_mask;
    
  if (!zone) return;

  printk(KERN_ERR "zone %08x: ", (unsigned int) zone);
  
  for (current_order = 0; current_order < MAX_ORDER; current_order++)
  {
    order_mask = (-1) << current_order;
    i = 0;
    area = zone->free_area + current_order;
    if (area->free_list.next == NULL)
    {
      printk("N ");
      continue;
    }
    spin_lock_irqsave(&zone->lock, flags);
    list_for_each_entry(page, &area->free_list, lru)
    { 
      i++;
      /*if (((badpage - page_to_pfn(page)) & order_mask) == 0)
	printk("*(%lx)", page_to_pfn(page));*/
    }
    spin_unlock_irqrestore(&zone->lock, flags);
    printk("%i(%li) ", i, area->nr_free);
  }
  
  printk("\n");
}

/* Given a pointed-at address and a mask, increment the page so that the
 * mask hides the increment. Return 0 if no increment is possible.
 */
static int __init next_masked_address (unsigned long *addrp, unsigned long mask){
        unsigned long inc=1;
        unsigned long newval = *addrp;
        while (inc & mask)
                inc += inc;
        while (inc != 0) {
                newval += inc;
                newval &= ~mask;
                newval |= ((*addrp) & mask);
                if (newval > *addrp) {
                        *addrp = newval;
                        return 1;
                }
                do {
                        inc += inc;
                } while (inc & ~mask);
                while (inc & mask)
                        inc += inc;
        }
        return 0;
}


void __init badram_markpages (int argc, unsigned long *argv) {
        unsigned long addr, mask;
        while (argc-- > 0) {
                addr = *argv++;
                mask = (argc-- > 0) ? *argv++ : ~0L;
                mask |= ~PAGE_MASK;     /* Optimalisation */
                addr &= mask;           /* Normalisation */
                do {
                        struct page *pg = phys_to_page(addr);
			printk(KERN_ERR "marking page %lx\n", 
				page_to_pfn(pg));
//                        printk ("%05lx ", __pa(__va(addr)) >> PAGE_SHIFT);
//                        printk ("=%05lx/%05lx ", (unsigned long)(pg-mem_map),
//                                        max_mapnr);
                        /* if (VALID_PAGE(pg)) {*/
			mark_page_bad(pg);
                        /* }*/
                        /* else printk ("INV ");*/
                } while (next_masked_address (&addr,mask));
        }
}

/*********** CONFIG_BADRAM: CUSTOMISABLE SECTION STARTS HERE ******************/

/* Enter your custom BadRAM patterns here as pairs of unsigned long integers. *//* For more information on these F/M pairs, refer to Documentation/badram.txt */

static unsigned long __initdata badram_custom[] = {
       0,      /* Number of longwords that follow, as F/M pairs */
};


/*********** CONFIG_BADRAM: CUSTOMISABLE SECTION ENDS HERE ********************/

char badram_string[1024];
module_param_string(badram, badram_string, 1024, 0);

static int __init badram_setup (char *str)
{
        unsigned long opts[3];
        if (!mem_map) BUG();
        printk (KERN_ERR "PAGE_OFFSET=0x%08lx\n", PAGE_OFFSET);
        printk (KERN_ERR "BadRAM option is %s\n", str);
//        if (*str++ == '=')
                while ((str = get_options (str, 3, (int *) opts), *opts)) {
                        printk (KERN_ERR 
                            "   --> marking 0x%08lx, 0x%08lx  [%ld]\n",
                                        opts[1], opts[2], opts[0]);
                        badram_markpages (*opts, opts+1);
                        if (*opts == 1)
                                break;
                };
        badram_markpages (*badram_custom, badram_custom+1);
        return 0;
}

static int __init userbadram_init_module(void)
{ 
  int i;

//  printk(KERN_ERR "badram=%s\n", badram_string);

  badram_setup(badram_string);

//  printk(KERN_ERR "\n");
/*  printk(KERN_ERR "max_mapnr:     %li\n", max_mapnr);
  printk(KERN_ERR "num_physpages: %li\n", num_physpages);
  printk(KERN_ERR "mem_map:       0x%lx\n", (unsigned long int) mem_map);*/
  
//    print_page(pfn_to_page(badpage), -1);
 
/*  printk(KERN_ERR "ZONES_SHIFT: %i\n", ZONES_SHIFT);
  printk(KERN_ERR "NODES_SHIFT: %i\n", NODES_SHIFT);*/

  if (0)
    for (i = 0; i < 1 << (ZONES_SHIFT + NODES_SHIFT); i++)
    {
      print_zone(zone_table[i]);
    }

//  mark_page_bad(pfn_to_page(badpage));
  
//  if (0) print_buddies(pfn_to_page(badpage));
  
  return 0;
} 

static void __exit userbadram_cleanup_module(void)
{
}

module_init(userbadram_init_module);
module_exit(userbadram_cleanup_module);
MODULE_LICENSE("GPL");
